<?php

declare(strict_types=1);

namespace mirzaev\vk\arangodb;

// Файлы проекта
use mirzaev\arangodb\collection;

// Фреймворк ArangoDB
use mirzaev\arangodb\document;
use mirzaev\vk\arangodb\longpoll;

// Библиотека для работы с API-сервера ArangoDB
use ArangoDBClient\Connection as _connection;
use ArangoDBClient\Statement as _statement;
use ArangoDBClient\Document as _document;

/**
 * Журнал
 *
 * Используется для хранения информации об изменениях других коллекций
 * Связывается с отслеживаемой коллекцией ребром
 *
 * @param _connection $session Сессия соединения с базой данных
 * @param _document $document Инстанция (static::COLLECTION_JOURNAL)
 */
class journal
{
    /**
     * Настройки архитектуры базы данных
     */
    const COLLECTION_JOURNAL = 'journal';

    /**
     * Конструктор
     *
     * @param _connection $session Сессия соединения с базой данных
     * @param _document $document Инстанция (static::COLLECTION_JOURNAL)
     */
    public function __construct(
        public _connection $session,
        public _document $document
    ) {
    }

    /**
     * Инициализация
     *
     * @param _connection $session Сессия соединения с базой данных
     * @param string $document Идентификатор документа
     * @param bool $create Создавать коллекции при их отсутствии
     *
     * @return static|null Объект с инстанцией журнала
     */
    public static function init(_connection $session, string $document, bool $create = true): ?static
    {
        if ($create) {
            // Запрошено создание коллекций в случае их отсутствия

            // Инициализация коллекций
            collection::init($session, static::COLLECTION_JOURNAL);
            collection::init($session, longpoll::COLLECTION_ACCESSED, true);
        }

        if (empty($journal = static::search($session, $document))) {
            // Не найден журнал (подразумевается, что его не существует)

            if (empty($journal = document::write($session, static::COLLECTION_JOURNAL, [
                'events' => [],
                'expires' => strtotime('first day of next month 00:00')
            ]))) {
                // Не удалось создать документ

                return null;
            }

            if (empty(document::write($session, longpoll::COLLECTION_ACCESSED, ['_from' => $document, '_to' => $journal]))) {
                // Не удалось создать ребро: {$document} -> ЖУРНАЛ

                return null;
            }
        }

        // Инициализация инстанции журнала и возврат
        return new static($session, $journal);
    }

    /**
     * Поиск журнала
     *
     * Находит актуальный документ журнала
     *
     * @param _connection $session Сессия соединения с базой данных
     * @param string $document Идентификатор документа
     *
     * @return _document|null Инстанция (static::COLLECTION_JOURNAL) или его идентификатор
     */
    public static function search(_connection $session, string $document): _document|null
    {
        // Инициализация ярлыка названия коллекции
        $collection = static::COLLECTION_JOURNAL;

        // Инициализация ярлыка названия ребра
        $edge = longpoll::COLLECTION_ACCESSED;

        // Инициализация даты для разделения по журналам
        $time = time();

        // Поиск журнала
        $journal = (new _statement(
            $session,
            [
                'query' => <<<AQL
                    FOR a IN $collection
                    LET b = (
                        FOR vertex, edge IN INBOUND a $edge
                        FILTER vertex._id == $document
                        LIMIT 1
                        RETURN vertex
                    )
                    FILTER a.expires >= $time
                    LIMIT 1
                    RETURN a
                AQL,
                "batchSize" => 1000,
                "sanitize"  => true
            ]
        ))->execute();

        // Инициализация буфера вывода
        $buffer = [];

        foreach ($journal as $key => $value) {
            // Перебор параметров для универсализации вывода

            // Запись в буфер
            $buffer[$key] = $value;
        }

        // Возврат (подразумевается, что в буфере только 1 значение - инстанция (static::COLLECTION_JOURNAL))
        return $buffer[0] ?? null;
    }

    /**
     * Запись в журнал
     *
     * @param string $event Событие
     * @param array $data Дополнительная информация
     *
     * @return bool Статус обработки
     */
    public function write(string $event, array $data = []): bool
    {
        // Запись в инстанцию документа
        $this->document->events += [
            [
                'date' => time(),
                'event' => $event,
                'data' => $data
            ]
        ];

        // Запись в базу данных
        return document::update($this->session, $this->document);
    }
}
